/** \file VectorOperations.h */

#ifndef VECTOR_OPERATIONS_H
#define VECTOR_OPERATIONS_H

#include "AMDiS.h"
#include "ValueTypes.h"

#if HAVE_PARALLEL_DOMAIN_AMDIS
#include "parallel/StdMpi.h"
#endif

#include <unistd.h>
#include <ios>
#include <iostream>
#include <fstream>
#include <string>
#include <boost/numeric/mtl/mtl.hpp>
#include <boost/numeric/itl/itl.hpp>
#include <boost/utility/enable_if.hpp>

using namespace AMDiS; 

template<typename T>
const WorldVector<T>& operator*=(WorldVector<T>& v1, const WorldVector<T>& v2)
{
  T* v1It, *v2It;
  for (v1It = v1.begin(), v2It = v2.begin();
	  v1It != v1.end(); 
	  v1It++, v2It++)
  *v1It *= *v2It;
  
  return v1;
};

// x ~= y
template<typename T>
class compareTol : public binary_function<T, T, bool> {
public:
	compareTol(double tol_) : tol(tol_) {}
	bool operator()(const T &x, const T &y) const { return sqrt((x-y)*(x-y))<tol; }
	double tol;
};
template<typename T1, typename T2, int component>
struct comparePair : public binary_function<std::pair<T1, T2>, std::pair<T1, T2>, bool> 
{
  bool operator()(const std::pair<T1, T2> &x, const std::pair<T1, T2> &y) const 
  { 
    return (component==0?(x.first<y.first):(x.second<y.second));
  }
};

template<typename T1, typename T2>
struct compare0 : public binary_function<std::pair<T1, T2>, std::pair<T1, T2>, bool> 
{
  bool operator()(const std::pair<T1, T2> &x, const std::pair<T1, T2> &y) const 
  { 
    return x.first<y.first;
  }
};
template<typename T1, typename T2>
struct compare1 : public binary_function<std::pair<T1, T2>, std::pair<T1, T2>, bool> 
{
  bool operator()(const std::pair<T1, T2> &x, const std::pair<T1, T2> &y) const 
  { 
    return x.second<y.second;
  }
};

class SortPair : public binary_function<std::pair<WorldVector<double>, double>, std::pair<WorldVector<double>, double>, bool> {
public:
	SortPair(int component_=1,int idx_=0) : component(component_),idx(idx_) {}
	bool operator()(const std::pair<WorldVector<double>, double> &x, 
		const std::pair<WorldVector<double>, double> &y) const { 
		return (component==1?(x.first)[idx]<(y.first)[idx]:x.second<y.second); }
	int component,idx;
};

namespace vector_operations {

  namespace traits {

    /// General declaration, used to disable unsupported types
    template <typename Collection, class Enable = void>
    struct num_rows {};
    
    /// size implementation for STL vectors
    template <typename Value>
    struct num_rows< std::vector<Value> > 
    {
      typedef std::size_t   type;
      type operator()(const std::vector<Value>& v) { return v.size(); }
    };

    /// size implementation for (1D) arrays interpreted as vectors
    template <typename Value, unsigned Size>
    struct num_rows<Value[Size]>
    {
      typedef std::size_t   type;
      type operator()(const Value[Size]) { return Size; }
    };	   

    /// size implementation for (2D and higher) arrays interpreted as matrices
    template <typename Value, unsigned Rows, unsigned Cols>
    struct num_rows<Value[Rows][Cols]>
    {
      typedef std::size_t   type;
      type operator()(const Value[Rows][Cols]) { return Rows; }
    };	    
    
    template <typename Value>
    struct num_rows< WorldVector<Value> > 
    {
      typedef std::size_t   type;
      type operator()(const WorldVector<Value>& v) { return static_cast<size_t>(v.getSize()); }
    };     
    
    template <typename Value>
    struct num_rows< WorldMatrix<Value> > 
    {
      typedef std::size_t   type;
      type operator()(const WorldMatrix<Value>& v) { return static_cast<size_t>(v.getNumRows()); }
    };   
    
    template <typename Value>
    struct num_rows< Value, typename boost::enable_if< boost::is_arithmetic<Value> >::type > 
    {
      typedef std::size_t   type;
      type operator()(const Value& v) { return 1; }
    };
    
    //________________________________________________________________________________    
    
    /// General declaration, used to disable unsupported types
    template <typename Collection, class Enable = void>
    struct num_cols {};
    
    template <typename Value>
    struct num_cols< WorldMatrix<Value> > 
    {
      typedef std::size_t   type;
      type operator()(const WorldMatrix<Value>& v) { return static_cast<size_t>(v.getNumCols()); }
    };   
    
    //________________________________________________________________________________    
    
    /// General declaration, used to disable unsupported types
    template <typename Collection, class Enable = void>
    struct resize {};
    
    template <typename Value>
    struct resize< WorldVector<Value> > 
    {
      typedef void   vector_void_type;
      void operator()(WorldVector<Value>& v, size_t r) { 
	TEST_EXIT(Global::getGeo(WORLD) == r)
	  ("WorldVectors can not be resized!\n");
      }
    };   
    
    template <typename Value>
    struct resize< std::vector<Value> > 
    {
      typedef void   vector_void_type;
      void operator()(std::vector<Value>& v, size_t r) { 
	v.resize(r);
      }
    };   
    
    template <typename Value>
    struct resize< mtl::dense_vector<Value> > 
    {
      typedef void   vector_void_type;
      void operator()(mtl::dense_vector<Value>& v, size_t r) { 
	v.change_dim(r);
      }
    };
        
    // _________________________________________________________________________________
        
    template <typename Value>
    struct resize< WorldMatrix<Value> > 
    {
      typedef void   matrix_void_type;
      void operator()(WorldMatrix<Value>& v, size_t r, size_t c) { 
	TEST_EXIT(Global::getGeo(WORLD) == r && Global::getGeo(WORLD) == c)
	  ("WorldMatrices can not be resized!\n");
      }
    };   
    
    template <typename Value, typename Param>
    struct resize< mtl::matrix::base_matrix<Value, Param> > 
    {
      typedef void   matrix_void_type;
      void operator()(mtl::matrix::base_matrix<Value, Param>& v, size_t r, size_t c) { 
	v.change_dim(r,c);
      }
    };
    
    template <typename Value>
    struct resize< mtl::matrix::dense2D<Value> > 
    {
      typedef void   matrix_void_type;
      void operator()(mtl::matrix::dense2D<Value>& v, size_t r, size_t c) { 
	v.change_dim(r,c);
      }
    };
    
    template <typename Value>
    struct resize< mtl::matrix::compressed2D<Value> > 
    {
      typedef void   matrix_void_type;
      void operator()(mtl::matrix::compressed2D<Value>& v, size_t r, size_t c) { 
	v.change_dim(r,c);
      }
    };
  }

  /// num_rows function for non-MTL types (uses implicit enable_if)
  template <typename Collection>
  typename traits::num_rows<Collection>::type 
  inline num_rows(const Collection& c)
  {  
    return traits::num_rows<Collection>()(c);
  }
  
  /// num_cols function for non-MTL types (uses implicit enable_if)
  template <typename Collection>
  typename traits::num_cols<Collection>::type 
  inline num_cols(const Collection& c)
  {  
    return traits::num_cols<Collection>()(c);
  }
  
  /// resize function for vectors
  template <typename Collection>
  typename traits::num_cols<Collection>::vector_void_type 
  inline resize(const Collection& c, size_t rows)
  {  
    traits::resize<Collection>()(c, rows);
  }
  
  /// resize function for matrices
  template <typename Collection>
  typename traits::num_cols<Collection>::matrix_void_type 
  inline resize(const Collection& c, size_t rows, size_t cols)
  {  
    traits::resize<Collection>()(c, rows, cols);
  }
    
  template<typename T>
  T norm(mtl::dense_vector<T> &b) { return two_norm(b); };
  
  template<typename T>
  T norm(std::vector<T> &b) 
  {
    T value = 0.0;
    for (size_t i = 0; i < b.size(); ++i)
      value += sqr(b[i]);
    return sqrt(value);
  };
  
  template<typename Matrix>
  double matrix_norm(Matrix &M)
  {
    typename Matrix::value_type asum;
    asum = 0.0;
    for (size_t i = 0; i < num_rows(M); ++i) {
      for (size_t j = 0; j < num_cols(M); ++j)
	asum += abs(M[i][j]);
    }
    return static_cast<double>(asum/(num_rows(M)*num_cols(M)));
  };
  
  // |x|_eps
  inline double epsNorm(const WorldVector<double> &x, const double &eps) { return std::max(AMDiS::norm(x),eps); };
  
  // 1/|x|_eps
  inline double epsNormInv(const WorldVector<double> &x, const double &eps) { return 1.0/epsNorm(x,eps); };
  
  // 1/[|x|_eps]^3
  inline double epsNormInv3(const WorldVector<double> &x, const double &eps)
  {
    double result = epsNorm(x,eps);
    result = 1.0/(result*sqr(result));  
    return result;
  };
  
  // y:=x/|x|_eps
  inline void epsNormalize(const WorldVector<double> &x, const double &eps, WorldVector<double> &y)
  {
    double nrmInv = epsNormInv(x,eps);
    for(int k=0; k<x.getSize(); k++) {
      y[k] = x[k] * nrmInv;
    }
  };

  template<typename Matrix>
  inline typename boost::enable_if< is_matrix<Matrix>, double >::type
  trace(Matrix &M)
  {
    TEST_EXIT(num_rows(M) == num_cols(M))
      ("Matrix dimensions must be equal!\n");
    typename Matrix::value_type sum;
    nullify(sum);
    for (size_t i = 0; i < num_rows(M); ++i) {
      sum += M[i][i];
    }
    return static_cast<double>(sum);
  }
  
  template<typename T>
  Vector<T> mult(const Vector<T>& v1, const Vector<T>& v2)
  {
    TEST_EXIT(v1.getSize()==v2.getSize())("Vectors must have the same size!\n");
    Vector<T> result(v1.getSize());
    Vector<T> *v1It, *v2It, *resultIt;
    for (v1It = v1.begin(), v2It = v2.begin(), resultIt = result.begin(); v1It != v1.end(); ++v1It, ++v2It, ++resultIt)
	*resultIt = *v1It * *v2It;
    return result;
  };
  
  inline WorldVector<double> mult(const mtl::dense2D<double> &A, const WorldVector<double> &x)
  {
    TEST_EXIT(num_rows(A)==x.getSize())("Falsche Dimensionen multipliziert!");
    WorldVector<double> result;
    for (unsigned i = 0; i < x.getSize(); ++i) {
      result[i] = 0.0;
      for (unsigned j = 0; j < x.getSize(); ++j)
	result[i] += A[i][j]*x[j];
    }
    return result;
  };
  
  template<typename ResultVector, typename Matrix, typename Vector1, typename Vector2>
  ResultVector Axpy(const Matrix &A, const Vector1 &x, const Vector2 &y)
  {
    size_t n = num_rows(x);
    TEST_EXIT(num_rows(A)==n && num_cols(A)==n && num_rows(y)==n)("Falsche Dimensionen multipliziert!");
    ResultVector result;
    resize(result, n);
    for (unsigned i = 0; i < n; ++i) {
      result[i] = y[i];
      for (unsigned j = 0; j < n; ++j)
	result[i] += A[i][j]*x[j];
    }
    return result;
  };

  inline void unitVector(WorldVector<double> &b, int &idx) {
    TEST_EXIT(idx>=0 && idx<b.getSize())("Unit-Vector component does not exist!\n");
    b.set(0.0);
    b[idx]=1.0;
  };
  
  inline WorldVector<double> *unitVector(int &idx) {
    WorldVector<double> *result = new WorldVector<double>;
    TEST_EXIT(idx>=0 && idx<result->size())("Unit-Vector component does not exist!\n");
    (*result)[idx]=1.0;
    return result;
  };

  // copy v -> w
  template<typename Vector1, typename Vector2>
  inline void copy(Vector1 &v, Vector2 &w)
  {
    for (size_t i = 0; i < std::min(num_rows(v), num_rows(w)); i++)
      w[i] = v[i];
  }
  
  // v3:=[v1, v2]
  template<typename Vector1, typename Vector2, typename VectorOut>
  inline void merge(Vector1 &v1, Vector2 &v2, VectorOut &v3)
  {
    resize(v3, num_rows(v1) + num_rows(v2));
    for (size_t i = 0; i < num_rows(v1); i++)
      v3[i] = v1[i];
    for (size_t j = 0; j < num_rows(v2); j++)
      v3[j + num_rows(v1)] = v2[j];
  }
  
  template<typename Vector>
  inline void getMin(const Vector &v, typename ValueType<Vector>::type &minVal, size_t &minIdx)
  {
    typedef typename ValueType<Vector>::type T;
    TEST_EXIT(num_rows(v) > 0)("getMin of empty vector!\n");
    
    minVal = v[0];
    minIdx = 0;
    for (size_t i = 1; i < num_rows(v); i++) {
      if (v[i] < minVal) {
	minVal = v[i];
	minIdx = i;
      }
    }
  }
  
  template<typename Vector>
  inline void getMax(const Vector &v, typename ValueType<Vector>::type &maxVal, size_t &maxIdx)
  {
    typedef typename ValueType<Vector>::type T;
    TEST_EXIT(num_rows(v) > 0)("getMax of empty vector!\n");
    
    maxVal = v[0];
    maxIdx = 0;
    for (size_t i = 1; i < num_rows(v); i++) {
      if (v[i] > maxVal) {
	maxVal = v[i];
	maxIdx = i;
      }
    }
  }

  template<typename Vector>
  inline typename ValueType<Vector>::type
  min(const Vector &vec)
  {
    typename ValueType<Vector>::type minVal;
    size_t minIdx;
    getMin(vec, minVal, minIdx);
    return minVal;
  }

  template<typename Vector>
  inline typename ValueType<Vector>::type
  max(const Vector &vec)
  {
    typename ValueType<Vector>::type maxVal;
    size_t maxIdx;
    getMax(vec, maxVal, maxIdx);
    return maxVal;
  }
  
  template<typename Vector>
  inline typename ValueType<Vector>::type
  absmin(const Vector &v)
  {
    typedef typename ValueType<Vector>::type T;
    TEST_EXIT(num_rows(v) > 0)("absmin of empty vector!\n");
    
    T minVal = abs(v[0]);
    for(size_t i = 1; i < num_rows(v); i++) {
      if (std::abs(v[i]) < minVal) {
	minVal = std::abs(v[i]);
      }
    }
    return minVal;
  }
  
  template<typename Vector>
  inline typename ValueType<Vector>::type
  absmax(const Vector &v)
  {
    typedef typename ValueType<Vector>::type T;
    TEST_EXIT(num_rows(v) > 0)("absmax of empty vector!\n");
    
    T maxVal = abs(v[0]);
    for(size_t i = 1; i < num_rows(v); i++) {
      if (std::abs(v[i]) > maxVal) {
	maxVal = std::abs(v[i]);
      }
    }
    return maxVal;
  }
  
  template<typename Vector>
  inline typename ValueType<Vector>::type
  sum(const Vector &v)
  {
    typedef typename ValueType<Vector>::type T;
    size_t N = num_rows(v);
    T value; nullify(value);
    if (N > 0) {
      value = v[0];
      for (size_t i = 1; i < N; ++i)
	value += v[i];
    }
    return value;
  }
  
  template<typename Vector>
  inline typename ValueType<Vector>::type
  mean(const Vector &v)
  {
    double N = static_cast<double>(num_rows(v));
    return sum(v) * (1.0 / N);
  }
  
  template<typename Vector>
  void getMean(const Vector &v, typename ValueType<Vector>::type &meanValue)
  {
    meanValue = mean(v);
  }
  
  template<typename T1, typename T2, typename Compare>
  void sort(std::vector<T1> &vec1, std::vector<T2> &vec2, Compare &comp)
  {
    std::vector<pair<T1,T2> > order(vec1.size());
    
    // create vector of pairs
    for (int i=0; i<vec1.size(); ++i)
      order[i] = make_pair(vec1[i], vec2[i]);
    
    // sort vector of pairs
    sort(order.begin(), order.end(), comp);
    
    // copy sorted pairs back to vectors
    for (int i=0; i<vec1.size(); ++i) {
      vec1[i] = order[i].first;
      vec2[i] = order[i].second;
    }
  };
  
  template<typename T1, typename T2, typename Compare>
  void sort(std::vector<std::pair<T1,T2> > &vec, Compare &comp)
  {
    std::sort(vec.begin(), vec.end(), comp);
  };
  
  inline void unique(std::vector<WorldVector<double> > &vec, double tol, std::vector<unsigned> &ind)
  {
    compareTol<WorldVector<double> > comp(tol);
    unsigned newVec=0;
    for(unsigned i=0; i<vec.size();++i) {
      bool inNew=false;
      for(unsigned j=0;j<newVec;++j) {
	inNew=inNew||comp(vec[i],vec[j]);
	if(inNew) break;
      }
      if(!inNew) {vec[newVec]=vec[i]; newVec++; ind.push_back(i);}
    }
    vec.erase(vec.begin()+newVec,vec.end());
  };
  
  template<typename T1, typename T2>
  void unique(std::vector<std::pair<T1,T2> > &vec, double tol, int uniqueBy=1)
  {
//     comparePair comp = comparePair(tol,uniqueBy);
//     unsigned newVec=0;
//     for(unsigned i=0; i<vec.size();++i) {
//       bool inNew=false;
//       for(unsigned j=0;j<newVec;++j) {
// 	inNew=inNew||comp(vec[i],vec[j]);
// 	if(inNew) break;
//       }
//       if(!inNew) {vec[newVec]=vec[i]; newVec++;}
//     }
//     vec.erase(vec.begin()+newVec,vec.end());
  };
  
  inline void filter(std::vector<WorldVector<double> > &x, std::vector<bool> &filter)
  {
    unsigned j=0;
    for(unsigned i=0; i<x.size(); ++i) {
      if(filter[i]) {
	x[j]=x[i];
	j++;
      }
    }
    x.erase(x.begin()+j,x.end());
  };
  
  inline void filter(std::vector<WorldVector<double> > &x, std::vector<int> &filter)
  {
    unsigned j=0;
    for(unsigned i=0; i<filter.size(); ++i) {
      x[j]=x[filter[i]];
      j++;
    }
    x.erase(x.begin()+j,x.end());
  };
  template<typename T, typename T1, typename T2>
  T dot(std::vector<T1> &vec1, std::vector<T2> &vec2)
  {
    unsigned N=vec1.size();
    T value = 0;
    if(N>0) {
      value = static_cast<T>(vec1[0])*static_cast<T>(vec2[0]);
      for(unsigned i=1; i<N; ++i) {
	value += static_cast<T>(vec1[i])*static_cast<T>(vec2[i]);
      }
    }
    return value;
  };
  template<typename T, typename T1, typename T2>
  T dot(mtl::dense_vector<T1> &vec1, WorldVector<T2> &vec2)
  {
      unsigned N=vec2.getSize();
      T value = 0;
      if(N>0) {
	  value = static_cast<T>(vec1[0])*static_cast<T>(vec2[0]);
	  for(unsigned i=1; i<N; ++i) {
	      value += static_cast<T>(vec1[i])*static_cast<T>(vec2[i]);
	  }
      }
      return value;
  };
  
  template<class T>
  static void fillMatrix(WorldMatrix<T> &m, mtl::dense2D<T> &m_in)
  {
    m.resize(num_rows(m_in),num_cols(m_in));
    for (size_t i = 0; i < num_rows(m_in); ++i) {
      for (size_t j = 0; j < num_cols(m_in); ++j) {
	m[i][j] = m_in(i,j);
      }
    }
  };
}

#endif // VECTOR_OPERATIONS_H
